;Tetris Clone. (BlitzBasic / Blitz3D)
;Stephen Ancell.
;Project Start: 26 February 2013.
;Project Release: 14 March 2013



;Some computer have intermittent graphics setup issues when loading games created with Blitz3D, 
;this patch pauses program execution for a few seconds to help prevent that from happening.
;Color 255, 255, 0
;Text (GraphicsWidth() / 2), (GraphicsHeight() / 2), "Please Wait A Few Seconds!", True, True
;Delay 5000




;Set screen resolution to maximum and setup double buffering.
Global gfxMode = CountGfxModes()
Graphics GfxModeWidth(gfxMode), GfxModeHeight(gfxMode)

;GFX modes under test.
;Graphics 800, 600
;Graphics 1024, 768
;Graphics 1152, 864
;Graphics 1280, 768
;Graphics 1280, 960
;Graphics 1280, 1024
;Graphics 1920, 1080

SetBuffer BackBuffer()


;Define a few global things.

;This stores the postion where the gamefield will be drawn, 
;the size of each square in the gamefield, 
;the amount of squares in the gamefield (columns / rows), 
;the size of the gamefield.
Global gameFieldX, gameFieldY
Global gameFieldCellWidth, gameFieldCellHeight
Global gameFieldColumnCount, gameFieldRowCount
Global gameFieldWidth, gameFieldHeight

;This 'is' the gamefield.
Dim gameField(0, 0)


;This stores the next shape in line that will come into play when the current shape reaches the bottom, 
;the current shape in play, 
;the current turn (rotation) of the current shape, 
;the previous rotation of the current shape so that it can be reverted back to, if the current shape won't fit when rotated, 
;the column and row coordinates of where the current shape will be drawn relative to the gamefield, 
;timing variables for controlling the fall-speed of the current shape, 
;the rate at which the game speeds up as it progresses (progression rate), 
;timing variables for controlling the direction speed of the current shape, when the player presses a direction key.
;timing variables for controlling the fall-speed of blocks above cleared rows, 
;timing variables for controlling how often the level increses to progress the current shape speed, 
;the current level number.
Global nextShape$
Global currentShape$
Global currentShapeTurn$
Global previousShapeTurn$
Global currentShapeX, currentShapeY
Global currentShapeTime, currentShapeTimeDelay
Global currentshapeTimeDelayProgressRate
Global inputTime, inputTimeDelay
Global rowClearTime, rowClearTimeDelay
Global levelTime, levelTimeDelay
Global levelNumber


;Stores the current state of gameplay (true / false), for selecting between play mode and gameover mode.
Global gameOn


Global gameWasQuit


;Enable the player to save current screen to an image file by pressing the [S] key.
Global screenSaveEnabled = False


;Set this to false if you don't want a grid appearing over the gamefield.
Global showGameFieldGrid = True


;This store the current score of the game that's being played, 
;the maximum amount of highscores and the highscores table.
Global currentScore
Global currentScoreMultiplier

Global highScoreShowTime, highScoreShowTimeDelay
Global showHighScores = False
Global highScoreCount = 10
Global highScoreEntryRequired = False
Global highScoreTextField$ = ""
Dim highScoreName$(highScoreCount)
Dim highScore(highScoreCount)


;Stores the game soundtrack.
Global musicLoop
Global musicOn = True




;Set to false so that the program starts up in gameover mode.
gameOn = False

;Initiate a first gamefield so that the program don't error when it starts in gameover mode.
Setup()

If musicOn
	LoadMusic()
EndIf

LoadHighScores()




;Main Loop, this will run until the [Escape] key is pressed.
Repeat
	;Switch to gameplay mode and reset the gamefield when [Space] bar is pressed.
	If KeyDown(57) = True And gameOn = False
		If highScoreEntryRequired = False
			gameOn = True
			gameWasQuit = False
			Setup()
		EndIf
	EndIf
	
	If HighScoreShowTimer() = True
		showHighScores = showHighScores Xor True
	EndIf
	
	;showHighScores = True
	
	;Run game.
	If gameOn = True
		UpdateGame()
	EndIf
	
	;Make the whole screen drawable.
	Viewport 0, 0, GraphicsWidth(), GraphicsHeight()
	
	;Set the main background colour and clear the off-screen backbuffer.
	ClsColor 75, 75, 75
	Cls
	
	;Draw everything in the game to the off-screen backbuffer.
	DrawGame()
	
	;Run and draw gameover mode over the previously played gamefield.
	If gameOn = False And showHighScores = False
		GameOver()
	EndIf
	
	If showHighScores = True And gameOn = False
		DrawHighScores()
	EndIf
	
	If highScoreEntryRequired = True
		InputHighScore()
	EndIf
	
	;SPAM! everything that was drawn to the backbuffer onto the screen (frontbuffer).
	Flip
	
	;Save screen to file if [S] key is pressed.
	If screenSaveEnabled = True And KeyHit(31) = True
		SaveScreenToFile()
	EndIf
Until KeyHit(1) = True

;Properly exit the program when the main loop finishes.
End




;Call this function whenever a new game needs to start, 
;this function also initiates some of the global variables that were defined at the beginning of this source code.
Function Setup()
	gameFieldCellWidth = (GraphicsWidth() / 100)
	gameFieldCellHeight = (GraphicsHeight() / 60)
	gameFieldColumnCount = 20
	gameFieldRowCount = 40
	gameFieldX = (GraphicsWidth() / 2) - ((gameFieldCellWidth * gameFieldColumnCount) / 2)
	gameFieldY = (GraphicsHeight() / 2) - ((gameFieldCellHeight * gameFieldRowCount) / 2)
	gameFieldWidth = (gameFieldCellWidth * gameFieldColumnCount)
	gameFieldHeight = (gameFieldCellHeight * gameFieldColumnCount)
	levelNumber = 1
	
	currentShapeTime = MilliSecs()
	currentShapeTimeDelay = 1000
	currentshapeTimeDelayProgressRate = (currentShapeTimeDelay / 30)
	
	inputTime = MilliSecs()
	inputTimeDelay = 50
	
	rowClearTime = MilliSecs()
	rowClearTimeDelay = 150
	
	levelTime = MilliSecs()
	levelTimeDelay = 60000
	
	highScoreShowTime = MilliSecs()
	highScoreShowTimeDelay = 7000
	
	highScoreEntryRequired = False
	highScoreTextField$ = ""
	
	SeedRnd MilliSecs()
	SetupgameField()
	ResetCurrentScore()
	
	SetNextShape()
	ChangeShape()
	
	previousShapeTurn$ = currentShapeTurn$
End Function




;Sets up and clears the gamefield with 0's and fills the Last Line with 1's 
;which are there to stop the shape falling through the bottom of the gamefield, 
;the last line is not visible to the player.
Function SetupgameField()
	Local column, row
	Dim	gameField(gameFieldColumnCount, gameFieldRowCount)
	
	For row = 0 To gameFieldRowCount
		For column = 0 To (gameFieldColumnCount - 1)
			gameField(column, row) = 0
			
			If row = gameFieldRowCount
				gameField(column, row) = 1
			EndIf
		Next
	Next
End Function




Function IncreaseLevel()
	currentShapeTimeDelay = currentShapeTimeDelay - (currentshapeTimeDelayProgressRate + levelNumber)
	
	If currentShapeTimeDelay < 50
		currentShapeTimeDelay = 50
	EndIf
	
	levelNumber = levelNumber + 1
End Function




;A basic timer that returns true when the system timer matches the (currentShapeTime + currentShapeTimeDelay) variables, 
;used for controlling the the fall speed of the current shape.
;Apply this to a variable or conditional statement block.
Function CurrentShapeTimer()
	Local flag = False
	
	If MilliSecs() >= (currentShapeTime + currentShapeTimeDelay)
		flag = True
		CurrentShapeTimerReset()
	EndIf
	
	Return flag
End Function




Function CurrentShapeTimerReset()
	currentShapeTime = MilliSecs()
End Function




;A basic timer that returns true when the system timer matches the (inputTime + inputTimeDelay) variables, 
;used for controlling the horizontal speed of the shape when controlled by the player.
;Apply this to a variable or conditional statement block.
Function InputTimer()
	Local flag = False
	
	If MilliSecs() >= (inputTime + inputTimeDelay)
		flag = True
		InputTimerReset()
	EndIf
	
	Return flag
End Function




Function InputTimerReset()
	inputTime = MilliSecs()
End Function




;A basic timer that returns true when the system timer matches the (rowClearTime + rowClearTimeDelay) variables, 
;used for controlling the fall-speed of blocks above cleared rows.
;Apply this to a variable or conditional statement block.
Function RowClearTimer()
	Local flag = False
	
	If rowClearTimeDelay > currentShapeTimeDelay
		rowClearTimeDelay = currentShapeTimeDelay
	EndIf
	
	If MilliSecs() >= (rowClearTime + rowClearTimeDelay)
		flag = True
		RowClearTimerReset()
	EndIf
	
	Return flag
End Function




Function RowClearTimerReset()
	rowClearTime = MilliSecs()
End Function




;A basic timer that returns true when the system timer matches the (levelTime + levelTimeDelay) variables, 
;used for controlling rate at which the shape speed progresses at each level.
;Apply this to a variable or conditional statement block.
Function LevelTimer()
	Local flag = False
	
	If MilliSecs() >= (levelTime + levelTimeDelay)
		flag = True
		LevelTimerReset()
	EndIf
	
	Return flag
End Function




Function LevelTimerReset()
	levelTime = MilliSecs()
End Function




;A basic timer that returns true when the system timer matches the (highScoreShowTime + highScoreShowTimeDelay) variables, 
;used for controlling the length of time the highscore table is displayed.
;Apply this to a variable or conditional statement block.
Function HighScoreShowTimer()
	Local flag = False
	
	If MilliSecs() >= (highScoreShowTime + highScoreShowTimeDelay)
		flag = True
		HighScoreShowTimerReset()
	EndIf
	
	Return flag
End Function




Function HighScoreShowTimerReset()
	highScoreShowTime = MilliSecs()
End Function




Function UpdateGame()
	If KeyHit(16) = True
		gameOn = False
		gameWasQuit = True
	EndIf
	
	;If squares or shapes are stacked all the way to the top of the gamefield then set gameover mode.
	If ShapeStackAtTop() = True
		gameOn = False
		
		If currentScore > highScore(highScoreCount - 1) And gameWasQuit = False
			highScoreEntryRequired = True
		EndIf
		
		FlushKeys()
	EndIf
	
	;Detect player input keys.
	PlayerInput()
	
	If LevelTimer() = True
		IncreaseLevel()
	EndIf
	
	;Go through the gamefield squares, find a completed line, 
	;clear that line and then move all lines above that line down.
	If RowClearTimer() = True
		FindCompleteGameFieldRows()
	EndIf
	
	;If the are no filled squares below any of the squares in the current shape, 
	;move the current shape down if CurrentShapeTimer() is true or if the player presses the down arrow key.
	If ShapeCanMoveDown(currentShape$, currentShapeTurn$, currentShapeX - TrimShapeLeft(currentShape$, currentShapeTurn$), currentShapeY - TrimShapeTop(currentShape$, currentShapeTurn$)) = True
		If CurrentShapeTimer() = True Or KeyDown(208) = True
			MoveCurrentShapeDown()
		EndIf
	EndIf
	
	;If the are any filled squares below any of the squares in the current shape, 
	;check that the currentShapeTimer is true so that extra movements side to side can be made, allows the current shape to do quick final tuck-ins into the nooks and crannies, 
	;copy the shape's squares into the gamefield array, giving the illusion that the shape has been fixed to the gamefield, 
	;swap the current shape for the next shape.
	If ShapeCanMoveDown(currentShape$, currentShapeTurn$, currentShapeX - TrimShapeLeft(currentShape$, currentShapeTurn$), currentShapeY - TrimShapeTop(currentShape$, currentShapeTurn$)) = False
		If CurrentShapeTimer() = True
			GlueShapeToGameField(currentShape$, currentShapeTurn$, currentShapeX - TrimShapeLeft(currentShape$, currentShapeTurn$), currentShapeY - TrimShapeTop(currentShape$, currentShapeTurn$))
			ChangeShape()
		EndIf
	EndIf
End Function




;Call this function when in gameover mode (gameOn = false)
Function GameOver()
	;Make the whole screen drawable.
	Viewport 0, 0, GraphicsWidth(), GraphicsHeight()
	
	;Displays some basic game instructions in the gameover window.
	ControlsInfo()
End Function




;Call this function from within the GameOver() function.
Function ControlsInfo()
	;Set the gameover window size.
	Local width = (StringWidth(" ") * 50)
	Local height = (StringHeight(" ") * 25)
	
	;Set the gameover window position centre screen.
	Local x = (GraphicsWidth() / 2) - (width / 2)
	Local y = (GraphicsHeight() / 2) - (height / 2)
	
	;Display text info in relevant colours, inside the gameover window.
	Color 50, 50, 50
	Rect x, y, width, height, True
	
	Color 255, 150, 255
	Text x + (width / 2), y + (StringHeight(" ") * 2), "Programmed With Blitz3D (BlitzBasic)", True, True
	Text x + (width / 2), y + (StringHeight(" ") * 4), "By Steve Ancell", True, True
	
	Color 255, 255, 0
	Text x + (width / 2), y + (StringHeight(" ") * 6), "G A M E  O V E R!", True, True
	
	Color 150, 150, 150
	Text x + (width / 2), y + (StringHeight(" ") * 10), "[Z] [X] Rotate Shape", True, True
	Text x + (width / 2), y + (StringHeight(" ") * 12), "[Left Arrow] [Right Arrow] Shape Left & Right", True, True
	Text x + (width / 2), y + (StringHeight(" ") * 14), "[Down Arrow] Speed Up Shape Drop", True, True
	
	Color Rnd(255), Rnd(255), Rnd(255)
	Text x + (width / 2), y + (StringHeight(" ") * 20), "Press [SPACE] To Start Game", True, True
	
	Color 150, 150, 150
	Text x + (width / 2), y + (StringHeight(" ") * 22), "[Q] Quit Game", True, True
	Text x + (width / 2), y + (StringHeight(" ") * 23), "[Escape] Exit", True, True
End Function




;Call this from within the UpdateGame() function, detects the keypresses required to play the game.
Function PlayerInput()
	;If [Z] key is pressed, rotate the current shape counter-clockwise.
	If KeyHit(44) = True
		TurnCurrentShapeLeft()
		
		;If the rotated shape won't fit, revert back to the previous rotation.
		If ShapeFits(currentShape$, currentShapeTurn$, currentShapeX - TrimShapeLeft(currentShape$, currentShapeTurn$), currentShapeY - TrimShapeTop(currentShape$, currentShapeTurn$)) = False
			currentShapeTurn$ = previousShapeTurn$
		EndIf
	EndIf
	
	;If [X] key is pressed, rotate the current shape clockwise.
	If KeyHit(45) = True
		TurnCurrentShapeRight()
		
		;If the rotated shape won't fit, revert back to the previous rotation.
		If ShapeFits(currentShape$, currentShapeTurn$, currentShapeX - TrimShapeLeft(currentShape$, currentShapeTurn$), currentShapeY - TrimShapeTop(currentShape$, currentShapeTurn$)) = False
			currentShapeTurn$ = previousShapeTurn$
		EndIf
	EndIf
	
	;If [Left Arrow] is pressed, check if any space exists between the current shape and the left side of the gamefield.
	If KeyDown(203) = True
		If ShapeAgainstLeftSide(currentShape$, currentShapeTurn$, currentShapeX - TrimShapeLeft(currentShape$, currentShapeTurn$)) = False
			;If there are no filled squares next to the left side of each square in the current shape, 
			;wait for the input timer and then move the current shape one square space to the left.
			If ShapeCanMoveLeft(currentShape$, currentShapeTurn$, currentShapeX - TrimShapeLeft(currentShape$, currentShapeTurn$), currentShapeY - TrimShapeTop(currentShape$, currentShapeTurn$)) = True
				If InputTimer() = True
					MoveCurrentShapeLeft()
				EndIf
			EndIf
		EndIf
	EndIf
	
	;If [Right Arrow] is pressed, check if any space exists between the current shape and the right side of the gamefield.
	If KeyDown(205) = True
		If ShapeAgainstRightSide(currentShape$, currentShapeTurn$, currentShapeX - TrimShapeLeft(currentShape$, currentShapeTurn$)) = False
			;If there are no filled squares next to the right side of each square in the current shape, 
			;wait for the input timer and then move the current shape one square space to the right.
			If ShapeCanMoveRight(currentShape$, currentShapeTurn$, currentShapeX - TrimShapeLeft(currentShape$, currentShapeTurn$), currentShapeY - TrimShapeTop(currentShape$, currentShapeTurn$)) = True
				If InputTimer() = True
					MoveCurrentShapeRight()
				EndIf
			EndIf
		EndIf
	EndIf
End Function




;Apply this function to a variable or conditional statement block to get the amount of rotations in a shape.
Function GetShapeTurnCount(shapeName$)
	Local shapeData$
	Local shapeTurnCount
	
	;Start at the beginning of the shapes datablock.
	Restore shapes
	
	;Name search loop.
	Repeat
		Read shapeData$
		
		If shapeData$ = shapeName
			Exit
		EndIf
	Forever
	
	;Rotation search loop.
	Repeat
		Read shapeData$
		
		If Instr(shapeData$, "TurnCount")
			shapeTurnCount = Mid(shapeData$, 11, Len(shapeData$) - 10)
			Exit
		EndIf
	Forever
	
	Return shapeTurnCount
End Function




;Apply this function to a variable or conditional statement block 
;to check if a shape is butted against the Left side of the gameField.
Function ShapeAgainstLeftSide(shapeName$, shapeTurn$, shapeX)
	Local shapeData$
	Local column, row
	Local flag = False
	
	SeekShape(shapeName, shapeTurn)
	
	For row = 0 To 4
		For column = 0 To 4
			Read shapeData$
			
			If shapeData$ = 1
				If (column + shapeX) = 0
					flag = True
					Exit
				EndIf
			EndIf
		Next
		
		If flag = True
			Exit
		EndIf
	Next
	
	Return flag
End Function




;Apply this function to a variable or conditional statement block 
;to check if a shape is butted against the Right side of the gameField.
Function ShapeAgainstRightSide(shapeName$, shapeTurn$, shapeX)
	Local shapeData$
	Local column, row
	Local flag = False
	
	SeekShape(shapeName, shapeTurn)
	
	For row = 0 To 4
		For column = 0 To 4
			Read shapeData$
			
			If shapeData$ = 1
				If (column + shapeX) = (gameFieldColumnCount - 1)
					flag = True
					Exit
				EndIf
			EndIf
		Next
		
		If flag = True
			Exit
		EndIf
	Next
	
	Return flag
End Function




;Apply this function to a variable or conditional statement block 
;to check if a shape is butted against the right side of a square that's already on the gamefield..
Function ShapeCanMoveLeft(shapeName$, shapeTurn$, shapeX, shapeY)
	Local shapeData$
	Local column, row
	Local gameFieldColumn, gameFieldRow
	Local flag = True
	
	SeekShape(shapeName, shapeTurn)
	
	For row = 0 To 4
		For column = 0 To 4
			Read shapeData$
			
			gameFieldColumn = (column + shapeX)
			gameFieldRow = (row + shapeY)
			
			If shapeData$ = 1
				If gameField((gameFieldColumn - 1), gameFieldRow) = 1
					flag = False
					Exit
				EndIf
			EndIf
		Next
		
		If flag = False
			Exit
		EndIf
	Next
	
	Return flag
End Function




;Apply this function to a variable or conditional statement block 
;to check if a shape is butted against the left side of a square that's already on the gamefield..
Function ShapeCanMoveRight(shapeName$, shapeTurn$, shapeX, shapeY)
	Local shapeData$
	Local column, row
	Local gameFieldColumn, gameFieldRow
	Local flag = True
	
	SeekShape(shapeName, shapeTurn)
	
	For row = 0 To 4
		For column = 0 To 4
			Read shapeData$
			
			gameFieldColumn = (column + shapeX)
			gameFieldRow = (row + shapeY)
			
			If shapeData$ = 1
				If gameField((gameFieldColumn + 1), gameFieldRow) = 1
					flag = False
					Exit
				EndIf
			EndIf
		Next
		
		If flag = False
			Exit
		EndIf
	Next
	
	Return flag
End Function




;Apply this function to a variable or conditional statement block 
;to check if any squares in a shape intersect with any squares already on the gamefield.
Function ShapeFits(shapeName$, shapeTurn$, shapeX, shapeY)
	Local shapeData$
	Local column, row
	Local gameFieldColumn, gameFieldRow
	Local flag = True
	
	SeekShape(shapeName, shapeTurn)
	
	For row = 0 To 4
		For column = 0 To 4
			Read shapeData$
			
			gameFieldColumn = (column + shapeX)
			gameFieldRow = (row + shapeY)
			
			If shapeData$ = 1
				If gameField(gameFieldColumn, gameFieldRow) = 1
					flag = False
					Exit
				EndIf
				
				If gameFieldColumn < 0 Or gameFieldColumn > (gameFieldColumnCount - 1)
					flag = False
					Exit
				EndIf
			EndIf
		Next
		
		If flag = False
			Exit
		EndIf
	Next
	
	Return flag
End Function




;Call this function whenever the current shape reaches the bottom of the gamefield.
Function ChangeShape()
	;Copy next shape into the current shape then change the next shape.
	SetCurrentShape()
	SetNextShape()
	
	;Reset the rotation and position of the current shape.
	;currentShape$ = "STest 1"
	currentShapeTurn$ = "Turn 1"
	currentShapeX = 8
	currentShapeY = 0
End Function




;Call this function from within ChangeShape() to change next shape to a new random shape.
Function SetNextShape()
	Local index = Rnd(1, 10)
	nextShape$ = "Shape " + index
End Function




;Call this function from within ChangeShape() to change current shape to next shape.
Function SetCurrentShape()
	currentShape$ = "Shape " + GetNextShapeIndex()
End Function




;Call this function to get the numerical value from the next shape name.
Function GetNextShapeIndex()
	Local index
	index = Mid(nextShape$, 7, Len(nextShape$) - 6)
	Return index
End Function




;Call this function to get the numeric value from the current shape name.
Function GetCurrentShapeIndex()
	Local index = Mid(currentShape$, 7, Len(currentShape$) - 6)
	Return index
End Function




;Call this function to set the rotation of the current shape, using a numeric value.
Function SetCurrentShapeTurn(turnIndex)
	currentShapeTurn$ = "Turn " + turnIndex
End Function




;Call this function to get the numeric value from the current shape turn name.
Function GetCurrentShapeTurnIndex()
	Local turnIndex = Mid(currentShapeTurn$, 6, Len(currentShapeTurn$) - 5)
	Return turnIndex
End Function




;Call this function to rotate the current shape counter-clockwise.
Function TurnCurrentShapeLeft()
	previousShapeTurn$ = currentShapeTurn$
	
	SetCurrentShapeTurn(GetCurrentShapeTurnIndex() - 1)
	
	If GetCurrentShapeTurnIndex() < 1
		SetCurrentShapeTurn(GetShapeTurnCount(currentShape$))
	EndIf
End Function




;Call this function to rotate the current shape clockwise.
Function TurnCurrentShapeRight()
	previousShapeTurn$ = currentShapeTurn$
	
	SetCurrentShapeTurn(GetCurrentShapeTurnIndex() + 1)
	
	If GetCurrentShapeTurnIndex() > GetShapeTurnCount(currentShape$)
		SetCurrentShapeTurn(1)
	EndIf
End Function




;Call this function to move the current shape one square space to the left.
Function MoveCurrentShapeLeft()
	currentShapeX = currentShapeX - 1
End Function




;Call this function to move the current shape one square space to the right.
Function MoveCurrentShapeRight()
	currentShapeX = currentShapeX + 1
End Function




;Call this function to move the current shape one square space down.
Function MoveCurrentShapeDown()
	currentShapeY = currentShapeY + 1
End Function




;Apply this function to a variable or conditional statement block 
;to check if any squares in the gamefield are directly below any of the squares in the parsed shape.
Function ShapeCanMoveDown(shapeName$, shapeTurn$, shapeX, shapeY)
	Local shapeData$
	Local column, row
	Local gameFieldColumn, gameFieldRow
	Local flag = True
	
	SeekShape(shapeName, shapeTurn)
	
	For row = 0 To 4
		For column = 0 To 4
			Read shapeData$
			
			gameFieldColumn = (column + shapeX)
			gameFieldRow = (row + shapeY)
			
			If shapeData$ = 1
				If gameField(gameFieldColumn, (gameFieldRow + 1)) = 1
					flag = False
					Exit
				EndIf
			EndIf
		Next
		
		If flag = False
			Exit
		EndIf
	Next
	
	Return flag
End Function




;Call this function whenever a shape reaches the bottom of the gamefield 
;or crashes into any squares below.
Function GlueShapeToGameField(shapeName$, shapeTurn$, shapeX, shapeY)
	Local shapeData$
	Local column, row
	Local gameFieldColumn, gameFieldRow
	Local flag = True
	
	SeekShape(shapeName, shapeTurn)
	
	For row = 0 To 4
		For column = 0 To 4
			Read shapeData$
			
			gameFieldColumn = (column + shapeX)
			gameFieldRow = (row + shapeY)
			
			If shapeData$ = 1
				gameField(gameFieldColumn, gameFieldRow) = 1
			EndIf
		Next
	Next
End Function




;Call this function to fast-forward the data pointer to a shape and rotation of that shape.
Function SeekShape(shapeName$, shapeTurn$)
	Local shapeData$
	Local shapeTurnCount$
	
	;Start at beginning of shape data block.
	Restore shapes
	
	Repeat
		Read shapeData$
		
		If shapeData = shapeName
			Exit
		EndIf
	Forever
	
	Repeat
		Read shapeData$
		
		If Instr(shapeData$, "TurnCount")
			shapeTurnCount$ = Mid(shapeData$, 11, Len(shapeData$) - 10)
			Exit
		EndIf
	Forever
	
	Repeat
		Read shapeData$
		
		If shapeData$ = shapeTurn
			Exit
		EndIf
	Forever
End Function




;Call this function to clear a completed row.
Function FindCompleteGameFieldRows()
	Local gameFieldColumn, gameFieldRow
	Local fillCount
	
	For gameFieldRow = 0 To (gameFieldRowCount - 1)
		fillCount = 0
		
		For gameFieldColumn = 0 To (gameFieldColumnCount - 1)
			If gameField(gameFieldColumn, gameFieldRow) = 1
				fillCount = fillCount + 1
			EndIf
			
			If fillCount = gameFieldColumnCount
				ClearGameFieldRow(gameFieldRow)
				IncreaseScore(currentScoreMultiplier)
				
				;Remove or comment-out the next commmand if you want this function to clear all completed rows in one pass, 
				;leave it as it is if you want to delete a completed row every frame so that the player can see the incomplete rows move down, block by block.
				Return
			EndIf
		Next
	Next
End Function




;Clears a gamefield row, 
;call from within the FindCompleteGameFieldRows() function.
Function ClearGameFieldRow(gameFieldRow)
	Local gameFieldColumn
	
	For gameFieldColumn = 0 To (gameFieldColumnCount - 1)
		gameField(gameFieldColumn, gameFieldRow) = 0
	Next
	
	RemoveClearedGameFieldRow(gameFieldRow)
End Function




;Moves all rows above the row parsed To this Function down one square space.
Function RemoveClearedGameFieldRow(clearedGameFieldRow)
	Local gameFieldColumn, gameFieldRow
	
	For gameFieldRow = clearedGameFieldRow To 0 Step -1
		For gameFieldColumn = 0 To (gameFieldColumnCount - 1)
			If gameFieldRow > 0
				gameField(gameFieldColumn, gameFieldRow) = gameField(gameFieldColumn, gameFieldRow - 1)
			EndIf
			
			If (gameFieldRow - 1) = 0
				gameField(gameFieldColumn, gameFieldRow) = 0
			EndIf
		Next
	Next
End Function




;Detects if there is a stack of squares that reach the top of the gamefield, 
;used for setting gameover mode.
Function ShapeStackAtTop()
	Local gameFieldColumn
	Local gameFieldRow = 0
	Local flag = False
	
	For gameFieldColumn = 0 To (gameFieldColumnCount - 1)
		If gameField(gameFieldColumn, gameFieldRow) = 1
			flag = True
			Exit
		EndIf
	Next
	
	Return flag
End Function




;Call from within Main Loop to draw everything in the game.
Function DrawGame()
	Viewport gameFieldX, gameFieldY, (gameFieldCellWidth * gameFieldColumnCount), (gameFieldCellHeight * gameFieldRowCount)
	
	ClsColor 100, 100, 100
	Cls
	DrawCurrentShape()
	DrawGameField()
	
	Viewport 0, 0, GraphicsWidth(), GraphicsHeight()
	
	DrawLevelNumber()
	DrawScore()
	DrawNextShape()
End Function




Function DrawNextShape()
	Local x = (gameFieldX + gameFieldWidth) + gameFieldCellWidth
	Local y = gameFieldY
	
	Color 255, 255, 0
	Text x, y, "NEXT SHAPE"
	
	Color 10, 10, 10
	DrawShapeAbsolute(nextShape$, "Turn 1", x, (y + StringHeight(" ")))
End Function




;Draw the current shape onto the gamefield at the stored position, 
;call from within the DrawGame() function.
Function DrawCurrentShape()
	Color 10, 10, 10
	DrawShape(currentShape$, currentShapeTurn$, currentShapeX - TrimShapeLeft(currentShape$, currentShapeTurn$), currentShapeY - TrimShapeTop(currentShape$, currentShapeTurn$))
End Function




;Draws a given shape, with a given rotation, at a given position on the gamefield.
Function DrawShape(shapeName$, shapeTurn$, shapeX, shapeY)
	Local shapeData$
	Local column, row
	Local cellX, cellY
	
	SeekShape(shapeName, shapeTurn)
	
	For row = 0 To 4
		For column = 0 To 4
			Read shapeData$
			
			If shapeData$ = "End"
				Exit
			EndIf
			
			cellX = (gameFieldCellWidth * column) + (gameFieldCellWidth * shapeX) + gameFieldX
			cellY = (gameFieldCellHeight * row) + (gameFieldCellHeight * shapeY) + gameFieldY
			
			;Color 10, 10, 10
			
			If shapeData$ = 1
				Rect cellX, cellY, gameFieldCellWidth, gameFieldCellHeight, True
			EndIf
		Next
		
		If shapeData$ = "End"
			Exit
		EndIf
	Next
End Function




Function DrawShapeAbsolute(shapeName$, shapeTurn$, absoluteX, absoluteY)
	Local shapeData$
	Local column, row
	Local cellX, cellY
	
	SeekShape(shapeName, shapeTurn)
	
	For row = 0 To 4
		For column = 0 To 4
			Read shapeData$
			
			If shapeData$ = "End"
				Exit
			EndIf
			
			cellX = (gameFieldCellWidth * column) + absoluteX
			cellY = (gameFieldCellHeight * row) + absoluteY
			
			;Color 10, 10, 10
			
			If shapeData$ = 1
				Rect cellX, cellY, gameFieldCellWidth, gameFieldCellHeight, True
			EndIf
		Next
		
		If shapeData$ = "End"
			Exit
		EndIf
	Next
End Function




;Subtract the result of this function from the vertical coordinate of a shape 
;to draw that shape with an offset that starts at the shape, rather than at the padding at the top of that shape.
Function TrimShapeTop(shapeName$, shapeTurn$)
	Local shapeData$
	Local column, row
	Local discardedTrim
	Local flag = False
	
	SeekShape(shapeName, shapeTurn)
	
	For row = 0 To 4
		For column = 0 To 4
			Read shapeData$
			
			If shapeData$ = 1
				discardedTrim = row
				flag = True
				
				Exit
			EndIf
		Next
		
		If flag = True
			Exit
		EndIf
	Next
	
	Return discardedTrim
End Function




;Subtract the result of this function from the horizontal coordinate of a shape 
;to draw that shape with an offset that starts at the shape, rather than at the padding to the left of that shape.
Function TrimShapeLeft(shapeName$, shapeTurn$)
	Local cell
	Local shapeData$
	Local column, row
	Local discardedTrim
	Local temp = CreateBank(25)
	Local flag = False
	
	SeekShape(shapeName, shapeTurn)
	
	For row = 0 To 4
		For column = 0 To 4
			Read shapeData$
			PokeByte temp, column + (row * 5), shapeData$
		Next
	Next
	
	For column = 0 To 4
		For row = 0 To 4
			cell = PeekByte(temp, column + (row * 5))
			
			If cell = 1
				discardedTrim = column
				flag = True
				
				Exit
			EndIf
		Next
		
		If flag = True
			Exit
		EndIf
	Next
	
	FreeBank temp
	Return discardedTrim
End Function




;Draw the gamefield, the squares on the gamefield and a grid over the gamefield, if the global variable showGameGrid is true.
;Call after drawing the current shape, if you want to see the shape as marked squares.
Function DrawGameField()
	Local column, row
	Local x, y
	
	If showGameFieldGrid = True
		For row = 0 To (gameFieldRowCount - 1)
			For column = 0 To (gameFieldColumnCount - 1)
				x = (gameFieldCellWidth * column) + gameFieldX
				y = (gameFieldCellHeight * row) + gameFieldY
				
				If gameField(column, row) = 1
					Color 10, 10, 10
					Rect x, y, gameFieldCellWidth, gameFieldCellHeight, True
				EndIf
				
				Color 50, 50, 50
				Rect x, y, gameFieldCellWidth, gameFieldCellHeight, False
			Next
		Next
	EndIf
	
	Color 150, 150, 150
	Rect gameFieldX, gameFieldY, (gameFieldCellWidth * gameFieldColumnCount), (gameFieldCellHeight * gameFieldRowCount), False
End Function




Function DrawLevelNumber()
	Local x = gameFieldX + (gameFieldWidth / 2)
	Local y = gameFieldY - (StringHeight(" ") * 1)
	
	Color 0, 255, 255
	Text x, y, "Level: " + levelNumber, True, True
End Function




Function DrawScore()
	Local x = gameFieldX + (gameFieldWidth / 2)
	Local y = gameFieldY - (StringHeight(" ") * 3)
	
	If screenSaveEnabled = True
		Color Rnd(255), Rnd(255), Rnd(255)
		Text x, y - (StringHeight("") * 2), "Press [S] key to save current screen.", True, True
	EndIf
	
	Color 0, 255, 255
	Text x, y, "Score: " + currentScore, True, True
End Function




Function IncreaseScore(scoreMultiplier)
	currentScore = currentScore + scoreMultiplier
End Function




Function ResetCurrentScore()
	currentScore = 0
	currentScoreMultiplier = gameFieldColumnCount
End Function




Function LoadHighScores()
	Local scoreIndex
	Local file = ReadFile("highscores.dat")
	
	If Not file
		ResetHighScores()
		file = ReadFile("highscores.dat")
	EndIf
	
	For scoreIndex = 0 To (highScoreCount - 1)
		highScoreName$(scoreIndex) = ReadLine(file)
		highScore(scoreIndex) = ReadLine(file)
	Next
	
	CloseFile file
End Function




Function SaveHighScores()
	DeleteFile("highscores.dat")
	
	Local scoreIndex
	Local file = WriteFile("highscores.dat")
	
	For scoreIndex = 0 To (highScoreCount - 1)
		WriteLine file, highScoreName$(scoreIndex)
		WriteLine file, highScore(scoreIndex)
	Next
	
	CloseFile file
End Function




Function InputHighScore()
	Local x = (GraphicsWidth() / 2)
	Local y = (GraphicsHeight() / 2)
	Local maxNameLength = 30
	Local char = GetKey()
	Local boxWidth = (StringWidth(" ") * maxNameLength)
	Local boxheight = (StringHeight(" ") * 5)
	Local noChars
	
	If Len(highScoreTextField$) < 1
		noChars = True
	EndIf
	
	If Len(highScoreTextField$) > highScore(highScoreCount)
		noChars = False
	EndIf
	
	If boxWidth < (StringWidth(" ") * 10)
		boxWidth = (StringWidth(" ") * 10)
	EndIf
	
	If KeyHit(28) = True Or KeyHit(156) = True
		If highScoreTextField$ = ""
			highScoreTextField$ = "Anonymous"
		EndIf
		
		highScoreEntryRequired = False
		SubmitHighScore()
	EndIf
	
	If char > 31 And char < 126
		If Len(highScoreTextField$) < maxNameLength
			highScoreTextField$ = highScoreTextField$ + Chr(char)
		EndIf
	EndIf
	
	If KeyHit(14) = True
		highScoreTextField$ = Mid(highScoreTextField$, 1, Len(highScoreTextField$) - 1)
	EndIf
	
	Color 0, 0, 0
	Rect x - (boxWidth / 2), y - (boxheight / 2), boxWidth, boxheight, True
	
	Color 0, 255, 0
	Text x, y, highScoreTextField$, True, True
	
	If noChars = True
		Text x, y, "Enter Name Here!", True, True
	EndIf
End Function




Function SubmitHighScore()
	Local scoreIndex
	Local entryIndex
	
	SortHighScores()
	
	For scoreIndex = 0 To (highScoreCount - 1)
		If currentScore > highScore(scoreIndex)
			entryIndex = scoreIndex
			Exit
		EndIf
	Next
	
	For scoreIndex = (highScoreCount - 1) To entryIndex Step -1
		If scoreIndex > 0
			highScoreName$(scoreIndex) = highScoreName$(scoreIndex - 1)
			highScore(scoreIndex) = highScore(scoreIndex - 1)
		EndIf
	Next
	
	highScoreName$(entryIndex) = highScoreTextField$
	highScore(entryIndex) = currentScore
	
	SaveHighScores()
End Function




Function SortHighScores()
	Local scoreIndex
	Local nestedScoreIndex
	
	Local highestScore
	Local highestScoreIndex
	
	For scoreIndex = 0 To (highScoreCount - 1)
		highestScore = highScore(scoreIndex)
		
		For nestedScoreIndex = scoreIndex To (highScoreCount - 1)
			If highScore(nestedScoreIndex) > highestScore
				highestScore = highScore(nestedScoreIndex)
				highestScoreIndex = nestedScoreIndex
			EndIf
		Next
		
		If highestScore > highScore(scoreIndex)
			SwapHighScores(scoreIndex, highestScoreIndex)
		EndIf
	Next
End Function




Function SwapHighScores(scoreIndex1, scoreIndex2)
	Local name1$ = highScoreName$(scoreIndex1)
	Local score1 = highScore(scoreIndex1)
	
	Local name2$ = highScoreName$(scoreIndex2)
	Local score2 = highScore(scoreIndex2)
	
	highScoreName$(scoreIndex1) = name2$
	highScore(scoreIndex1) = score2
	
	highScoreName$(scoreIndex2) = name1$
	highScore(scoreIndex2) = score1
End Function




Function DrawHighScores()
	Local scoreIndex
	Local scoreName$, score
	
	;Set size of the highscores window.
	Local width = (StringWidth(" ") * 50)
	Local height = StringHeight(" ") * (highScoreCount * 2) + StringHeight(" ")
	
	;Set the highscore window position centre screen.
	Local x = (GraphicsWidth() / 2) - (width / 2)
	Local y = (GraphicsHeight() / 2) - (height / 2)
	
	Color 50, 50, 50
	Rect x, y, width, height, True
	
	For scoreIndex = 0 To (highScoreCount - 1)
		scoreName$ = (highScoreName$(scoreIndex) + " ")
		score = highScore(scoreIndex)
		
		Color 0, 255, 0
		Text x + (width / 2), y + ((StringHeight(" ") * scoreIndex) * 2) + StringHeight(" "), scoreName$ + score, True, False
	Next
End Function




Function ResetHighScores()
	Local scoreIndex
	Local file = WriteFile("highscores.dat")
	
	For scoreIndex = 0 To highScoreCount
		If scoreIndex < highScoreCount
			WriteLine file, "NO NAME HERE YET!"
			WriteLine file, 0
		EndIf
	Next
	
	CloseFile file
End Function




Function LoadMusic()
	musicLoop = LoadSound("media\1989 tetris theme tune.ogg")
	LoopSound musicLoop
	PlaySound musicLoop
End Function




Function SaveScreenToFile()
	Local fileName$ = "ScreenGrab"
	
	If gameOn = True
		SaveBuffer(FrontBuffer(), fileName$ + ".bmp")
	EndIf
	
	If gameOn = False
		SaveBuffer(FrontBuffer(), fileName$ + "Menu.bmp")
	EndIf
End Function




;Data blocks. Contains a data description that describes the segments of each available shape.
;0 = Empty segment.
;1 = Filled segment.
.shapes
;Shape 1
Data "Shape 1"
Data "TurnCount 2"
Data "Turn 1"
Data 0, 0, 0, 0, 0
Data 0, 0, 0, 0, 0
Data 1, 1, 1, 1, 1
Data 0, 0, 0, 0, 0
Data 0, 0, 0, 0, 0
Data "End"

Data "Turn 2"
Data 0, 0, 1, 0, 0
Data 0, 0, 1, 0, 0
Data 0, 0, 1, 0, 0
Data 0, 0, 1, 0, 0
Data 0, 0, 1, 0, 0
Data "End"




;Shape 2
Data "Shape 2"
Data "TurnCount 4"
Data "Turn 1"
Data 0, 0, 0, 0, 0
Data 0, 0, 0, 0, 0
Data 1, 1, 1, 1, 1
Data 0, 0, 0, 0, 1
Data 0, 0, 0, 0, 0
Data "End"

Data "Turn 2"
Data 0, 0, 1, 0, 0
Data 0, 0, 1, 0, 0
Data 0, 0, 1, 0, 0
Data 0, 0, 1, 0, 0
Data 0, 1, 1, 0, 0
Data "End"

Data "Turn 3"
Data 0, 0, 0, 0, 0
Data 1, 0, 0, 0, 0
Data 1, 1, 1, 1, 1
Data 0, 0, 0, 0, 0
Data 0, 0, 0, 0, 0
Data "End"

Data "Turn 4"
Data 0, 0, 1, 1, 0
Data 0, 0, 1, 0, 0
Data 0, 0, 1, 0, 0
Data 0, 0, 1, 0, 0
Data 0, 0, 1, 0, 0
Data "End"




;Shape 3
Data "Shape 3"
Data "TurnCount 4"
Data "Turn 1"
Data 0, 0, 0, 0, 0
Data 0, 0, 0, 0, 0
Data 1, 1, 1, 1, 1
Data 1, 0, 0, 0, 0
Data 0, 0, 0, 0, 0
Data "End"

Data "Turn 2"
Data 0, 1, 1, 0, 0
Data 0, 0, 1, 0, 0
Data 0, 0, 1, 0, 0
Data 0, 0, 1, 0, 0
Data 0, 0, 1, 0, 0
Data "End"

Data "Turn 3"
Data 0, 0, 0, 0, 0
Data 0, 0, 0, 0, 1
Data 1, 1, 1, 1, 1
Data 0, 0, 0, 0, 0
Data 0, 0, 0, 0, 0
Data "End"

Data "Turn 4"
Data 0, 0, 1, 0, 0
Data 0, 0, 1, 0, 0
Data 0, 0, 1, 0, 0
Data 0, 0, 1, 0, 0
Data 0, 0, 1, 1, 0
Data "End"




;Shape 4
Data "Shape 4"
Data "TurnCount 1"
Data "Turn 1"
Data 0, 0, 0, 0, 0
Data 0, 1, 1, 1, 0
Data 0, 1, 1, 1, 0
Data 0, 1, 1, 1, 0
Data 0, 0, 0, 0, 0
Data "End"




;Shape 5
Data "Shape 5"
Data "TurnCount 2"
Data "Turn 1"
Data 0, 0, 0, 0, 0
Data 0, 1, 1, 0, 0
Data 0, 0, 1, 1, 0
Data 0, 0, 0, 0, 0
Data 0, 0, 0, 0, 0
Data "End"

Data "Turn 2"
Data 0, 0, 0, 0, 0
Data 0, 0, 1, 0, 0
Data 0, 1, 1, 0, 0
Data 0, 1, 0, 0, 0
Data 0, 0, 0, 0, 0
Data "End"




;Shape 6
Data "Shape 6"
Data "TurnCount 2"
Data "Turn 1"
Data 0, 0, 0, 0, 0
Data 0, 0, 1, 1, 0
Data 0, 1, 1, 0, 0
Data 0, 0, 0, 0, 0
Data 0, 0, 0, 0, 0
Data "End"

Data "Turn 2"
Data 0, 0, 0, 0, 0
Data 0, 1, 0, 0, 0
Data 0, 1, 1, 0, 0
Data 0, 0, 1, 0, 0
Data 0, 0, 0, 0, 0
Data "End"




;Shape 7
Data "Shape 7"
Data "TurnCount 4"
Data "Turn 1"
Data 0, 0, 0, 0, 0
Data 0, 0, 1, 0, 0
Data 0, 1, 1, 1, 0
Data 0, 0, 0, 0, 0
Data 0, 0, 0, 0, 0
Data "End"

Data "Turn 2"
Data 0, 0, 0, 0, 0
Data 0, 0, 1, 0, 0
Data 0, 0, 1, 1, 0
Data 0, 0, 1, 0, 0
Data 0, 0, 0, 0, 0
Data "End"

Data "Turn 3"
Data 0, 0, 0, 0, 0
Data 0, 0, 0, 0, 0
Data 0, 1, 1, 1, 0
Data 0, 0, 1, 0, 0
Data 0, 0, 0, 0, 0
Data "End"

Data "Turn 4"
Data 0, 0, 0, 0, 0
Data 0, 0, 1, 0, 0
Data 0, 1, 1, 0, 0
Data 0, 0, 1, 0, 0
Data 0, 0, 0, 0, 0
Data "End"




;Shape 8
Data "Shape 8"
Data "TurnCount 1"
Data "Turn 1"
Data 0, 0, 0, 0, 0
Data 0, 0, 1, 0, 0
Data 0, 1, 1, 1, 0
Data 0, 0, 1, 0, 0
Data 0, 0, 0, 0, 0
Data "End"




;Shape 9
Data "Shape 9"
Data "TurnCount 4"
Data "Turn 1"
Data 0, 0, 0, 0, 0
Data 0, 0, 0, 0, 0
Data 0, 1, 1, 1, 0
Data 0, 0, 0, 1, 0
Data 0, 0, 0, 0, 0
Data "End"

Data "Turn 2"
Data 0, 0, 0, 0, 0
Data 0, 0, 1, 0, 0
Data 0, 0, 1, 0, 0
Data 0, 1, 1, 0, 0
Data 0, 0, 0, 0, 0
Data "End"

Data "Turn 3"
Data 0, 0, 0, 0, 0
Data 0, 1, 0, 0, 0
Data 0, 1, 1, 1, 0
Data 0, 0, 0, 0, 0
Data 0, 0, 0, 0, 0
Data "End"

Data "Turn 4"
Data 0, 0, 0, 0, 0
Data 0, 0, 1, 1, 0
Data 0, 0, 1, 0, 0
Data 0, 0, 1, 0, 0
Data 0, 0, 0, 0, 0
Data "End"




;Shape 10
Data "Shape 10"
Data "TurnCount 4"
Data "Turn 1"
Data 0, 0, 0, 0, 0
Data 0, 0, 0, 0, 0
Data 0, 1, 1, 1, 0
Data 0, 1, 0, 0, 0
Data 0, 0, 0, 0, 0
Data "End"

Data "Turn 2"
Data 0, 0, 0, 0, 0
Data 0, 1, 1, 0, 0
Data 0, 0, 1, 0, 0
Data 0, 0, 1, 0, 0
Data 0, 0, 0, 0, 0
Data "End"

Data "Turn 3"
Data 0, 0, 0, 0, 0
Data 0, 0, 0, 1, 0
Data 0, 1, 1, 1, 0
Data 0, 0, 0, 0, 0
Data 0, 0, 0, 0, 0
Data "End"

Data "Turn 4"
Data 0, 0, 0, 0, 0
Data 0, 0, 1, 0, 0
Data 0, 0, 1, 0, 0
Data 0, 0, 1, 1, 0
Data 0, 0, 0, 0, 0
Data "End"




;Test Shape
Data "STest 1"
Data "TurnCount 1"
Data "Turn 1"
Data 1, 1, 1, 1, 1
Data 1, 1, 1, 1, 1
Data 1, 1, 1, 1, 1
Data 1, 1, 1, 1, 1
Data 1, 1, 1, 1, 1
;~IDEal Editor Parameters:
;~C#Blitz3D